@@ -1,5 +1,5 @@
 /*
-** Copyright 1998 - 2016 Double Precision, Inc.
+** Copyright 1998 - 2016 S. Varshavchik.
 ** See COPYING for distribution information.
 */
 
@@ -208,10 +208,11 @@ int login_callback(struct authinfo *ainfo, void *dummy)
 			{
 				alarm(0);
 
-				fprintf(stderr, "INFO: LOGIN, user=%s, ip=[%s], port=[%s], protocol=%s\n",
+				fprintf(stderr, "INFO: LOGIN, user=%s, ip=[%s], port=[%s], protocol=%s%s\n",
 					ainfo->address, safe_getenv("TCPREMOTEIP"),
 					safe_getenv("TCPREMOTEPORT"),
-					safe_getenv("PROTOCOL"));
+					safe_getenv("PROTOCOL"),
+					(p=getenv("IMAP_TLS")) != 0 && atoi(p) ? ", starttls=1" : "");
 
 				proxyloop(fd);
 				exit(0);
@@ -481,7 +482,7 @@ int main(int argc, char **argv)
 	writes("* OK [CAPABILITY ");
 	imapcapability();
 	writes("] Courier-IMAP ready. "
-	       "Copyright 1998-2021 Double Precision, Inc.  "
+	       "Copyright 1998-2022 S. Varshavchik.  "
 	       "See COPYING for distribution information.\r\n");
 	fprintf(stderr, "DEBUG: Connection, ip=[%s], port=[%s]\n", ip, port);
 	writeflush();
