@@ -1,5 +1,5 @@
 /*
-** Copyright 1998 - 2018 Double Precision, Inc.
+** Copyright 1998 - 2018 S. Varshavchik.
 ** See COPYING for distribution information.
 **
 */
@@ -28,15 +28,9 @@
 #if	HAVE_SYS_STAT_H
 #include	<sys/stat.h>
 #endif
-#if TIME_WITH_SYS_TIME
-#include <sys/time.h>
-#include <time.h>
-#else
+#include	<time.h>
 #if HAVE_SYS_TIME_H
-#include <sys/time.h>
-#else
-#include <time.h>
-#endif
+#include	<sys/time.h>
 #endif
 #include	<stdio.h>
 #include	<stdlib.h>
@@ -60,6 +54,14 @@
 #define POP3DLIST "courierpop3dsizelist"
 #define LISTVERSION 3
 
+#ifndef RUNTIME_START
+#define RUNTIME_START time(NULL)
+#endif
+
+#ifndef RUNTIME_CUR
+#define RUNTIME_CUR time(NULL)
+#endif
+
 extern void pop3dcapa();
 extern void pop3dlang(const char *);
 
@@ -190,7 +192,7 @@ static struct msglist **readpop3dlist(unsigned long *uid)
 	int vernum=0;
 	unsigned long size;
 
-	uidv=time(NULL);
+	uidv=RUNTIME_CUR;
 
 	if (fp == NULL ||
 	    fgets(linebuf, sizeof(linebuf)-1, fp) == NULL ||
@@ -299,6 +301,9 @@ static int savepop3dlist(struct msglist **a, size_t cnt,
 
 	struct maildir_tmpcreate_info createInfo;
 
+#ifdef SAVEHOOK
+	SAVEHOOK();
+#endif
 	maildir_tmpcreate_init(&createInfo);
 
 	createInfo.uniq="pop3";
@@ -488,7 +493,9 @@ static void sortmsgs()
 	}
 
 	if (prev_list[n])
+	{
 		savesizes=1;
+	}
 
 	for (i=0; prev_list[i]; i++)
 	{
@@ -663,7 +670,7 @@ char	buf[NUMBUFSIZE];
 	"This E-mail message was determined to be Unicode-formatted\n"	\
 	"but your E-mail reader does not support Unicode E-mail.\n\n"	\
 	"Please use an E-mail reader that supports POP3 with UTF-8\n"	\
-	"(see https://tools.ietf.org/html/rfc6856.html).\n\n"		\
+	"(see https://datatracker.ietf.org/doc/html/rfc6856).\n\n"		\
 	"This can also happen when the sender's E-mail program does not\n" \
 	"correctly format the sent message.\n\n"			\
 	"The original message is included as a separate attachment\n"	\
@@ -728,7 +735,7 @@ static void do_retr(unsigned i, unsigned *lptr)
 
 	if (msglist_a[i]->isutf8 && !utf8_enabled)
 	{
-		sprintf(boundary, "=_%d-%d", (int)getpid(), (int)time(NULL));
+		sprintf(boundary, "=_%d-%d", (int)getpid(), (int)RUNTIME_CUR);
 
 		sprintf(wrapheader, MIMEWRAPTXT, boundary, boundary, boundary,
 			mime_message_type);
@@ -910,7 +917,7 @@ static void acctout(const char *disc)
 
 	libmail_str_size_t(top_count, num1);
 	libmail_str_size_t(retr_count, num2);
-	libmail_str_time_t(time(NULL)-start_time, num3);
+	libmail_str_time_t(RUNTIME_CUR-start_time, num3);
 	libmail_str_size_t(bytes_received_count, numAR);
 	libmail_str_size_t(bytes_sent_count, numAS);
 
@@ -945,13 +952,10 @@ static void acctout(const char *disc)
 	free(p);
 }
 
-static RETSIGTYPE bye(int signum)
+static void bye(int signum)
 {
 	acctout("INFO: TIMEOUT");
 	exit(0);
-#if	RETSIGTYPE != void
-	return (0);
-#endif
 }
 
 static void loop()
@@ -1121,11 +1125,10 @@ char	*n;
 int main(int argc, char **argv)
 {
 char	*p;
-
 #ifdef HAVE_SETVBUF_IOLBF
 	setvbuf(stderr, NULL, _IOLBF, BUFSIZ);
 #endif
-	time(&start_time);
+	start_time=RUNTIME_START;
 
 	if ((authaddr=getenv("AUTHADDR")) == NULL ||
 	    *authaddr == 0)
@@ -1194,10 +1197,11 @@ char	*p;
 		exit(1);
 	}
 
-	fprintf(stderr, "INFO: LOGIN, user=%s, ip=[%s], port=[%s]\n",
+	fprintf(stderr, "INFO: LOGIN, user=%s, ip=[%s], port=[%s]%s\n",
 			authaddr,
 					remoteip,
-					remoteport);
+					remoteport,
+					(p=getenv("POP3_TLS")) != 0 && atoi(p) ? ", stls=1" : "");
 	fflush(stderr);
 
 	msglist_cnt=0;
