@@ -1,5 +1,5 @@
 /*
-** Copyright 1998 - 2011 Double Precision, Inc.
+** Copyright 1998 - 2011 S. Varshavchik.
 ** See COPYING for distribution information.
 */
 
@@ -82,7 +82,6 @@ static int harvest_records(struct rfc1035_res *res,
 	int q_type, int *found, int autoquery, int port);
 
 #define	HARVEST_AUTOQUERY	1
-#define	HARVEST_NODUPE		2
 
 static int add_arecords(struct rfc1035_res *res, struct rfc1035_mxlist **list,
 			struct rfc1035_reply *mxreply,
@@ -91,6 +90,7 @@ static int add_arecords(struct rfc1035_res *res, struct rfc1035_mxlist **list,
 {
 #if	RFC1035_IPV6
 	struct in6_addr	in;
+	int rc2;
 	int first_a=RFC1035_TYPE_A;
 	int second_a=RFC1035_TYPE_AAAA;
 	const char *prefer_ipv6=getenv("ESMTP_PREFER_IPV6_MX");
@@ -132,31 +132,53 @@ records weren't found.
 		second_a=RFC1035_TYPE_A;
 	}
 
+	/* Read cached records, for the less-preferred address type */
+
 	if (mxreply && !(opts & RFC1035_MX_QUERYALL))
 	{
-		if ((rc=harvest_records(res, list, mxreply, mxpreference,
-			mxname, second_a, &found, HARVEST_NODUPE,
-				port))
-		    != RFC1035_MX_OK)
-			return (rc);
+		rc2=harvest_records(res, list, mxreply, mxpreference,
+				    mxname, second_a, &found, 0,
+				    port);
 
-		if ((rc=harvest_records(res, list, mxreply, mxpreference,
-			mxname, first_a, &found, 0, port))
-				!= RFC1035_MX_OK)
-			return (rc);
-
-		if (found)	return (RFC1035_MX_OK);
+		if (rc2 != RFC1035_MX_OK && rc2 != RFC1035_MX_SOFTERR)
+			return rc2;
 	}
 
-	if ((rc=harvest_records(res, list, mxreply, mxpreference, mxname,
-		second_a, &found, HARVEST_AUTOQUERY|HARVEST_NODUPE, port))
-						!= RFC1035_MX_OK)
+	/* If there's nothing cached make an explicit query */
+
+	rc2=harvest_records(res, list, mxreply, mxpreference, mxname,
+			    second_a, &found, HARVEST_AUTOQUERY,
+			    port);
+
+	if (rc2 != RFC1035_MX_OK && rc2 != RFC1035_MX_SOFTERR)
+		return rc2;
+
+	/* Read cached records, for the preferred address type */
+
+	if (mxreply && !(opts & RFC1035_MX_QUERYALL))
+	{
+		rc=harvest_records(res, list, mxreply, mxpreference,
+				   mxname, first_a, &found, 0,
+				   port);
+
+		if (rc != RFC1035_MX_OK && rc != RFC1035_MX_SOFTERR)
+			return (rc);
+
+		if (rc == RFC1035_MX_SOFTERR && rc2 == RFC1035_MX_SOFTERR)
+			return rc;
+	}
+
+	/* If there's nothing cached make an explicit query */
+
+	rc=harvest_records(res, list, mxreply, mxpreference, mxname,
+			   first_a, &found, HARVEST_AUTOQUERY,
+			   port);
+
+	if (rc != RFC1035_MX_OK && rc != RFC1035_MX_SOFTERR)
 		return (rc);
 
-	if ((rc=harvest_records(res, list, mxreply, mxpreference, mxname,
-		first_a, &found, HARVEST_AUTOQUERY, port))
-			!= RFC1035_MX_OK)
-		return (rc);
+	if (rc == RFC1035_MX_SOFTERR && rc2 == RFC1035_MX_SOFTERR)
+		return rc;
 
 	if (found)	return (RFC1035_MX_OK);
 
@@ -179,18 +201,35 @@ static int harvest_records(struct rfc1035_res *res,
 	int q_type, int *found,
 	int flags, int port)
 {
-struct rfc1035_reply *areply=0;
-int index;
+	char	lookup_name[RFC1035_MAXNAMESIZE+1];
+
+	struct rfc1035_reply *areply=0;
+	int index;
 #if	RFC1035_IPV6
-struct in6_addr in;
+	struct in6_addr in;
 #else
-struct in_addr in;
+	struct in_addr in;
 #endif
 
+	/*
+	** Make a copy of mxname, because resolve_cname modifies it.
+	** That is rather rude, since harvest_records gets called multiple
+	** times.
+	**
+	** We still need to know what resolve_cname() did, since
+	** after resolve_cname() we call replysearch_all(), which needs to
+	** have the same hostname.
+	**
+	** mxname always points to a char[RFC1035_MAXNAMESIZE_1], so what's
+	** good for the goose is good for the gander.
+	*/
+
+	strcpy(lookup_name, mxname);
+
 	index= -1;
 
 	if (!mxreply || (
-		((index=rfc1035_replysearch_all( res, mxreply, mxname,
+		((index=rfc1035_replysearch_all( res, mxreply, lookup_name,
 					q_type,
 					RFC1035_CLASS_IN,
 					0)) < 0 ||
@@ -198,7 +237,7 @@ struct in_addr in;
 		&& (flags & HARVEST_AUTOQUERY))
 		)
 	{
-		index=rfc1035_resolve_cname(res, mxname,
+		index=rfc1035_resolve_cname(res, lookup_name,
 			q_type,
 			RFC1035_CLASS_IN, &areply, RFC1035_X_RANDOMIZE);
 		if (index < 0)
@@ -223,10 +262,11 @@ struct in_addr in;
 	}
 
 	for ( ; index >= 0 ;
-			index=rfc1035_replysearch_all( res, mxreply, mxname,
-					q_type,
-					RFC1035_CLASS_IN,
-					index+1))
+			index=rfc1035_replysearch_all( res, mxreply,
+						       lookup_name,
+						       q_type,
+						       RFC1035_CLASS_IN,
+						       index+1))
 	{
 		if (mxreply->allrrs[index]->rrtype != q_type)
 			continue;
@@ -234,13 +274,20 @@ struct in_addr in;
 #if RFC1035_IPV6
 		if (q_type == RFC1035_TYPE_A)
 		{
-		struct rfc1035_mxlist *q;
-
 			/* Map it to an IPv4 address */
 
 			rfc1035_ipv4to6(&in,
 				&mxreply->allrrs[index]->rr.inaddr);
 
+		}
+		else
+		{
+			in=mxreply->allrrs[index]->rr.in6addr;
+		}
+
+		{
+			struct rfc1035_mxlist *q;
+
 			/* See if it's already here */
 
 			for (q= *list; q; q=q->next)
@@ -252,13 +299,12 @@ struct in_addr in;
 				memcpy(&sin6, &q->address, sizeof(sin6));
 
 				if (memcmp(&sin6.sin6_addr, &in, sizeof(in))
-					== 0 && q->priority == mxpreference)
+					== 0 && q->priority == mxpreference &&
+				    strcmp(q->hostname, mxname) == 0)
 					break;
 			}
-			if ((flags & HARVEST_NODUPE) && q)	continue;
+			if (q)	continue;
 		}
-		else
-			in=mxreply->allrrs[index]->rr.in6addr;
 #else
 		in.s_addr=mxreply->allrrs[index]->rr.inaddr.s_addr;
 #endif
@@ -352,6 +398,12 @@ int seen_good=0;
 			replyp->allrrs[index]->rr.mx.mx_label, mxname) == 0)
 			continue;
 
+		if (*mxname == 0) /* MX . */
+		{
+			rfc1035_replyfree(replyp);
+			return (RFC1035_MX_NONE);
+		}
+
 		switch (add_arecords(res, list, replyp,
 				     replyp->allrrs[index]->rr.mx.preference,
 				     mxname,
