@@ -1,5 +1,5 @@
 /*
-** Copyright 2004-2011 Double Precision, Inc.
+** Copyright 2004-2016 S. Varshavchik.
 ** See COPYING for distribution information.
 */
 
@@ -12,16 +12,11 @@
 #include	<string.h>
 #include	<errno.h>
 #include	<sys/types.h>
-#if TIME_WITH_SYS_TIME
-#include	<sys/time.h>
 #include	<time.h>
-#else
 #if HAVE_SYS_TIME_H
 #include	<sys/time.h>
-#else
-#include	<time.h>
-#endif
 #endif
+#include	<idn2.h>
 
 struct rfc1035_spf_info {
 	const char *mailfrom;
@@ -674,7 +669,7 @@ static char do_ptr(const char *name,
 	if (rfc1035_aton(info->tcpremoteip, &pinfo.addr) < 0)
 	{
 		set_err_msg(info->errmsg_buf, info->errmsg_buf_size,
-			    "Invalid tcpremoteip.\n");
+			    "Invalid tcpremoteip.");
 		return SPF_FAIL;
 	}
 
@@ -692,7 +687,7 @@ static char do_ptr(const char *name,
 	if (pinfo.error)
 	{
 		set_err_msg(info->errmsg_buf, info->errmsg_buf_size,
-			    "ptr lookup failed.\n");
+			    "ptr lookup failed.");
 		return SPF_UNKNOWN;
 	}
 	return SPF_FAIL;
@@ -717,7 +712,7 @@ static char do_ipcheck(const char *name, struct rfc1035_spf_info *info,
 	if (rfc1035_aton(info->tcpremoteip, &addr) < 0)
 	{
 		set_err_msg(info->errmsg_buf, info->errmsg_buf_size,
-			    "Invalid tcpremoteip.\n");
+			    "Invalid tcpremoteip.");
 		return SPF_FAIL;
 	}
 
@@ -845,7 +840,7 @@ static char mechanism(const char *name,
 		{
 			free(domain_spec);
 			set_err_msg(info->errmsg_buf, info->errmsg_buf_size,
-				    "Invalid tcpremoteip.\n");
+				    "Invalid tcpremoteip.");
 			return SPF_FAIL;
 		}
 
@@ -859,7 +854,7 @@ static char mechanism(const char *name,
 		if (rc != 0)
 		{
 			set_err_msg(info->errmsg_buf, info->errmsg_buf_size,
-				    "IP address lookup failed.\n");
+				    "IP address lookup failed.");
 			return SPF_UNKNOWN;
 		}
 
@@ -887,9 +882,9 @@ static char mechanism(const char *name,
 		/*
 		** This mechanism matches if the <sending-host> is one of the
 		** MX hosts for a domain name.
-   
+
 		** MX = "mx" [ ":" domain-spec ] [ dual-cidr-length ]
-    
+
 		** SPF clients first perform an MX lookup on the <target-name>.
 		** SPF clients then perform an A lookup on each MX name
 		** returned, in order of MX priority.  The <sending-host> is
@@ -906,7 +901,7 @@ static char mechanism(const char *name,
 		{
 			free(domain_spec);
 			set_err_msg(info->errmsg_buf, info->errmsg_buf_size,
-				    "Invalid tcpremoteip.\n");
+				    "Invalid tcpremoteip.");
 			return SPF_FAIL;
 		}
 
@@ -914,11 +909,12 @@ static char mechanism(const char *name,
 					   domain_spec, RFC1035_MX_QUERYALL,
 					   &mxlist);
 		free(domain_spec);
-		if (rc)
+
+		if (rc && rc != RFC1035_MX_HARDERR)
 		{
 			rfc1035_mxlist_free(mxlist);
 			set_err_msg(info->errmsg_buf, info->errmsg_buf_size,
-				    "DNS MX lookup failed.\n");
+				    "DNS MX lookup failed.");
 			return SPF_ERROR;
 		}
 
@@ -1044,7 +1040,7 @@ static void setexp(const char *exp,
 	{
 		set_err_msg(info->errmsg_buf,
 			    info->errmsg_buf_size,
-			    "A DNS lookup error occured while"
+			    "A DNS lookup error occurred while"
 			    " fetching the SPF explanation record.");
 	}
 	else
@@ -1068,7 +1064,7 @@ static char lookup(struct rfc1035_spf_info *info)
 	char c;
 
 	/*
-	** 
+	**
 	** If a loop is detected, or if more than 20 subqueries are triggered,
 	** an SPF client MAY abort the lookup and return the result "unknown".
 	*/
@@ -1312,7 +1308,68 @@ static int do_expand(const char *str, struct rfc1035_spf_info *info,
 static char *expandc(const char *ipaddr);
 static char *expandi(const char *ipaddr);
 
+static char *get_macro2(struct rfc1035_spf_info *info, char name);
+
 static char *get_macro(struct rfc1035_spf_info *info, char name)
+{
+	char *v=get_macro2(info, name);
+	char *p;
+	char *buf;
+	char *newbuf;
+
+	if (!v)
+		return v;
+
+	switch (name) {
+	case 's':
+		/* s = responsible-sender */
+
+		p=strrchr(v, '@');
+
+		/* Find domain, convert to ACE */
+		if (!p || idn2_to_ascii_8z(++p, &buf, 0) != IDNA_SUCCESS)
+			break;
+
+		/* Buffer for local part + ACE domain */
+
+		newbuf=malloc(p-v+strlen(buf)+1);
+
+		if (!newbuf)
+		{
+			free(buf);
+			break;
+		}
+
+		/* Rebuild address with an ACE domain */
+
+		memcpy(newbuf, v, p-v);
+		strcpy(newbuf+(p-v), buf);
+		free(v);
+		v=newbuf;
+		free(buf);
+		break;
+	case 'o':
+		/* o = responsible-domain */
+	case 'd':
+		/* d = current-domain */
+	case 'p':
+		/* p = SMTP client domain name */
+	case 'h':
+		/* h = HELO/EHLO domain */
+	case 'r':
+		/* r = receiving domain */
+
+		if (idna_to_ascii_8z(v, &buf, 0) == IDNA_SUCCESS)
+		{
+			free(v);
+			v=buf;
+		}
+	}
+
+	return v;
+}
+
+static char *get_macro2(struct rfc1035_spf_info *info, char name)
 {
 	char *p;
 	const char *cp;
@@ -1442,7 +1499,7 @@ static unsigned tsplit(char *macro, char delimiter, char **wordptr)
 
 	}
 	return cnt;
-}		
+}
 
 static char *transform(char *macro,
 		       unsigned transformer_count,
