@@ -1,5 +1,5 @@
 /*
-** Copyright 2000-2016 Double Precision, Inc.
+** Copyright 2000-2016 S. Varshavchik.
 ** See COPYING for distribution information.
 */
 #include	"config.h"
@@ -22,7 +22,7 @@
 #include	<stdlib.h>
 #include	<ctype.h>
 #include	<netdb.h>
-#include	<idna.h>
+#include	<idn2.h>
 #if HAVE_DIRENT_H
 #include <dirent.h>
 #define NAMLEN(dirent) strlen((dirent)->d_name)
@@ -59,7 +59,7 @@
 
 struct proto_ops {
     char *n;
-    const SSL_METHOD * (*m)();
+    const SSL_METHOD * (*m)(void);
     int o;
 };
 
@@ -67,11 +67,15 @@ struct proto_ops {
 #define SSL_OP_NO_RENEGOTIATION 0
 #endif
 
+#ifndef SSL_OP_CIPHER_SERVER_PREFERENCE
+#define SSL_OP_CIPHER_SERVER_PREFERENCE 0
+#endif
+
 struct proto_ops op_list[] =
 {
 #ifdef SSL_OP_NO_TLSv1
 #ifdef SSL_OP_NO_TLSv1_1
-    { "TLSv1.2++", &SSLv23_method,  SSL_OP_ALL|SSL_OP_NO_SSLv2|SSL_OP_NO_SSLv3|SSL_OP_NO_TLSv1|SSL_OP_NO_TLSv1_1|SSL_OP_NO_RENEGOTIATION },
+    { "TLSv1.2++", &SSLv23_method,  SSL_OP_ALL|SSL_OP_NO_SSLv2|SSL_OP_NO_SSLv3|SSL_OP_NO_TLSv1|SSL_OP_NO_TLSv1_1|SSL_OP_NO_RENEGOTIATION|SSL_OP_CIPHER_SERVER_PREFERENCE},
     { "TLSv1.2+",  &SSLv23_method,  SSL_OP_ALL|SSL_OP_NO_SSLv2|SSL_OP_NO_SSLv3|SSL_OP_NO_TLSv1|SSL_OP_NO_TLSv1_1 },
     { "TLSv1.2",   &SSLv23_method,  SSL_OP_ALL|SSL_OP_NO_SSLv2|SSL_OP_NO_SSLv3|SSL_OP_NO_TLSv1|SSL_OP_NO_TLSv1_1 },
 #endif
@@ -201,11 +205,11 @@ static int hostmatch(const struct tls_info *info, const char *domain)
 		}
 	}
 
-	if (idna_to_unicode_8z8z(verify_domain, &idn_domain1, 0)
+	if (idn2_to_unicode_8z8z(verify_domain, &idn_domain1, 0)
 	    != IDNA_SUCCESS)
 		idn_domain1=0;
 
-	if (idna_to_unicode_8z8z(p, &idn_domain2, 0)
+	if (idn2_to_unicode_8z8z(p, &idn_domain2, 0)
 	    != IDNA_SUCCESS)
 		idn_domain2=0;
 
@@ -405,19 +409,50 @@ static void load_dh_params(SSL_CTX *ctx, const char *filename,
 	const struct tls_info *info=SSL_CTX_get_app_data(ctx);
 
 	BIO	*bio;
-	DH	*dh;
 
 	if (*cert_file_flags)
 		return;
 
+	ERR_clear_error();
+
 	if ((bio=BIO_new_file(filename, "r")) != 0)
 	{
+#if HAVE_PEM_READ_BIO_PARAMETERS_EX
+
+		OSSL_LIB_CTX *libctx=OSSL_LIB_CTX_get0_global_default();
+
+		EVP_PKEY *pkey=PEM_read_bio_Parameters_ex(bio, NULL, libctx,
+							  NULL);
+
+		if (pkey)
+		{
+			if (EVP_PKEY_is_a(pkey, "DH"))
+			{
+				if (SSL_CTX_set0_tmp_dh_pkey(ctx, pkey))
+				{
+					*cert_file_flags = 1;
+				}
+				else
+				{
+					EVP_PKEY_free(pkey);
+				}
+			}
+			else
+			{
+				EVP_PKEY_free(pkey);
+			}
+		}
+
+#else
+		DH	*dh;
+
 		if ((dh=PEM_read_bio_DHparams(bio, NULL, NULL, NULL)) != 0)
 		{
 			SSL_CTX_set_tmp_dh(ctx, dh);
 			*cert_file_flags = 1;
 			DH_free(dh);
 		}
+#endif
 		else
 		{
 			/*
@@ -427,7 +462,12 @@ static void load_dh_params(SSL_CTX *ctx, const char *filename,
 
 			int err=ERR_peek_last_error();
 
-			if (ERR_GET_LIB(err) == ERR_LIB_PEM
+			if (err == 0)
+				/*
+				** No error, and no parameters. Fail.
+				*/
+				;
+			else if (ERR_GET_LIB(err) == ERR_LIB_PEM
 			    && ERR_GET_REASON(err) == PEM_R_NO_START_LINE)
 			{
 				ERR_clear_error();
@@ -993,6 +1033,8 @@ SSL_CTX *tls_create_int(int isserver, const struct tls_info *info,
 			SSLeay_add_ssl_algorithms();
 #endif
 
+#if OPENSSL_VERSION_MAJOR < 3
+
 			while (RAND_status() != 1)
 			{
 				const char *p=random128();
@@ -1000,6 +1042,7 @@ SSL_CTX *tls_create_int(int isserver, const struct tls_info *info,
 
 				RAND_add(p, l, l/16);
 			}
+#endif
 		}
 	}
 
@@ -1207,7 +1250,11 @@ void tls_destroy(SSL_CTX *ctx)
 {
 	struct tls_info *info=SSL_CTX_get_app_data(ctx);
 
-	SSL_CTX_flush_sessions(ctx, 0); /* OpenSSL bug, 2002-08-07 */
+#if HAVE_SSL_CTX_FLUSH_SESSIONS_EX
+	SSL_CTX_flush_sessions_ex(ctx, 0);
+#else
+	SSL_CTX_flush_sessions(ctx, 0);
+#endif
 
 	SSL_CTX_free(ctx);
 
@@ -1268,7 +1315,12 @@ static int cache_add(SSL *ssl, SSL_SESSION *sess)
 	struct tls_info *info=SSL_get_app_data(ssl);
 	unsigned char buffer[BUFSIZ];
 	unsigned char *ucp;
-	time_t timeout= (time_t)SSL_SESSION_get_time(sess)
+	time_t timeout=
+#if HAVE_SSL_SESSION_GET_TIME_EX
+		SSL_SESSION_get_time_ex(sess)
+#else
+		(time_t)SSL_SESSION_get_time(sess)
+#endif
 		+ SSL_SESSION_get_timeout(sess);
 	unsigned int session_id_len;
 	void *session_id;
@@ -1522,7 +1574,7 @@ SSL *tls_connect(SSL_CTX *ctx, int fd)
 
 		if (info->peer_verify_domain)
 		{
-			if (idna_to_unicode_8z8z(info->peer_verify_domain,
+			if (idn2_to_unicode_8z8z(info->peer_verify_domain,
 						 &idn_domain1, 0)
 			    != IDNA_SUCCESS)
 				idn_domain1=0;
